# -*- coding: utf-8 -*-
import datetime
import reversion
from typing import Any, Dict, Optional, Tuple
from urllib3 import HTTPResponse

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.core.cache import cache
from django.db import models, transaction
from django.db.models import Q, F
from django.utils.html import strip_tags
from django.utils.translation import gettext_lazy as _

from tinymce.models import HTMLField
from notifications.push_notifications.client import SNSClient
from notifications.utils import get_tenant_identifier

User = get_user_model()


class NotificationError(Exception):
    ...


class NotificationAbstractModel(models.Model):
    class NotificationChoices(models.TextChoices):
        SIGN_COMPANY = 'SIGN_COMPANY'
        SIMPLE_OK = 'SIMPLE_OK'
        RELEASE_NOTES = 'RELEASE_NOTES'

    class TypeChoices(models.IntegerChoices):
        WEB = 0, _('Web notification')
        PUSH = 1, _('Push notification')
        EMAIL = 2, _('Email notification (not implemented)')

    name = models.CharField(
        max_length=250,
        unique=True,
        help_text="Used to reference notification, shown in title for some looks",
    )
    message = HTMLField(
        help_text="Full message as shown to user",
    )
    expires = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        help_text="Notification will not be shown after this time.",
    )
    attachment = models.FileField(
        default=None,
        blank=True,
        null=True,
    )
    active_from = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
    )
    needs_approval = models.BooleanField(
        default=False,
        help_text="Set this field if approval is necessary, see snooze time.",
    )
    snooze_lock = models.IntegerField(
        default=None,
        null=True,
        blank=True,
    )
    snooze_time = models.IntegerField(
        default=None,
        null=True,
        blank=True,
        help_text="If user dismisses message (when relevant), message is shown again after this many days.",
    )  # Days
    send_email = models.BooleanField(
        default=False,
    )
    # SIGN_COMPANY set as default because it's the only one current existing, safe to remove
    look = models.CharField(
        max_length=50,
        choices=NotificationChoices.choices,
        default=NotificationChoices.SIMPLE_OK,
        help_text="This controls the appearance of the notification.",
    )
    notification_type = models.IntegerField(
        _('Notification type'),
        choices=TypeChoices.choices,
        default=TypeChoices.WEB
    )
    image = models.ImageField(
        upload_to='notification_imgs',
        null=True,
        blank=True,
        help_text="Image to accompany notification (optional)",
    )
    display_only_if_url_path_matches_regex = models.CharField(
        max_length=64,
        default='.*',
        null=False,
        blank=False,
        help_text='Only display this notification if the provided regex matches the url-path',
    )
    aws_sns_id = models.UUIDField(
        _('Id of notification in AWS SNS'),
        default=None,
        null=True,
        blank=True,
        help_text=_('Message id in AWS, if notification is a push notification')
    )

    class Meta:
        verbose_name = _('Notification')
        verbose_name_plural = _('Notifications')
        abstract = True

    def __str__(self) -> str:
        return self.name

    @property
    def is_web(self) -> bool:
        return self.notification_type == self.TypeChoices.WEB

    @property
    def is_push(self) -> bool:
        return self.notification_type == self.TypeChoices.PUSH

    def save(self, *args, **kwargs) -> None:
        if self.is_push:
            self.message = strip_tags(self.message)
            self.message_is = strip_tags(self.message_is)
            self.message_en = strip_tags(self.message_en)
        return super().save(*args, **kwargs)

@reversion.register()
class Notification(NotificationAbstractModel):
    external_id = models.IntegerField(
        null=True,
        help_text="Used to reference externally created notifications",
    )
    recipients = models.ManyToManyField(
        User,
        through='UserNotification',
        default=None,
        blank=True,
    )
    groups = models.ManyToManyField(
        Group,
        default=None,
        blank=True,
    )
    topic = models.ForeignKey(
        'NotificationTopic',
        related_name='sent_notifications',
        default=None,
        null=True,
        blank=True,
        on_delete=models.SET_NULL
    )

    @classmethod
    def notification_key(cls, user: User) -> str:
        """Get notification key as it is referred to in cache

        Args:
            user (User): User instance

        Returns:
            str: Cache key
        """
        return f'notifications_unseen:{user.id}'

    @classmethod
    def unseen(cls, user: User) -> dict:
        """Get unseen notifications for user

        Args:
            user (User): User instance

        Returns:
            dict: Unseen notifications
        """
        notifications = cache.get(cls.notification_key(user), default={})
        for user_notif_id in notifications:
            # To make sure cache is up to date for existing notifications
            # before cookies are populated.
            upd_notif = cache.get(cls.notification_key(user), default={})
            if (user_notif := UserNotification.objects.filter(
                id=user_notif_id,
                notification_type=NotificationAbstractModel.TypeChoices.WEB
            ).first()) is not None:
                user_notif.update_unseen_cache(unseen=upd_notif)

        updated_notifications = cache.get(cls.notification_key(user), default={})
        return updated_notifications

    def create_usernotifications_for_groups(self, database_name: Optional[str] = None) -> None:
        """Create user notifications for groups

        Args:
            database_name (Optional[str], optional): Database name to use. Defaults to None.
        """
        for group in self.groups.all():
            for user in group.user_set.all():
                if database_name is not None:
                    UserNotification.objects.using(database_name).get_or_create(notification=self, user=user)
                else:
                    UserNotification.objects.get_or_create(notification=self, user=user)

    def send_push_notifications(self, default_language: str = None) -> None:
        """Send push notifications to all

        Args:
            default_language (str, optional): Default language to fallback on. Defaults to None.
        """
        if not self.is_push:
            return
        for un in UserNotification.objects.filter(
            notification=self
        ).annotate(language=F('user__profile__prefs__language')):
            un.send_push_notification(un.language or default_language)

    def save(self, *args, **kwargs):
        """Override save so it creates user notifications after saving"""
        super(Notification, self).save(*args, **kwargs)
        using = kwargs.get('using', None)
        if self.groups.exists():
            self.create_usernotifications_for_groups(database_name=using)
        if self.topic is not None:
            self.topic.send(self)


@reversion.register()
class UserNotification(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    notification = models.ForeignKey(Notification, on_delete=models.CASCADE)
    seen = models.BooleanField(default=False)
    timestamp = models.DateTimeField(auto_now_add=True)
    next_display = models.DateTimeField(default=None, null=True, blank=True)
    answer = models.BooleanField(default=None, null=True, blank=True)
    answer_string = models.CharField(max_length=255, default=None, blank=True, null=True)

    class Meta:
        verbose_name = _('User notification')
        verbose_name_plural = _('User notifications')

    def __str__(self) -> str:
        return "{} - {} - seen:{} - answer:{} - answer_string:{}".format(
            self.notification.name,
            self.user.username,
            self.seen,
            self.answer,
            self.answer_string
        )

    def update_unseen_cache(self, unseen: Optional[dict] = None) -> Any:
        """Updates unseen notifications in cache

        Args:
            unseen (Optional[dict], optional): Unseen items. Defaults to None.

        Returns:
            Any: Return value from cache methods
        """
        if unseen is None:
            unseen = Notification.unseen(self.user)

        if (self.seen and self.answer) or (
            self.notification.expires is not None and self.notification.expires < datetime.datetime.now()
        ):
            if self.id in unseen:
                del unseen[self.id]
        elif self.next_display:
            unseen[self.id] = self.next_display
        else:
            # 10 Minute in the past to make up for discrepancy between client time and server time
            unseen[self.id] = datetime.datetime.now() - datetime.timedelta(minutes=10)
        if len(unseen) == 0:
            return cache.delete(Notification.notification_key(self.user))
        return cache.set(Notification.notification_key(self.user), unseen, timeout=None)

    def send_push_notification(self, language: str = None) -> None:
        """Send push notification for this notification

        Args:
            language (str, optional): Language for notification. Defaults to None.
        """
        if language is None:
            profile = getattr(self.user, 'profile', None)
            prefs = getattr(profile, 'prefs', {})
            language = prefs.get('language', language)
        for device in self.user.devices.filter(active=True):
            try:
                device.send(self.notification, language)
            except NotificationError:
                pass

    def save(self, *args, **kwargs) -> None:
        """Override save so it updates cache and does some snoozing"""
        lang = kwargs.pop('default_language', 'is')
        if self.answer is not None:
            self.seen = True
        if self.seen and not self.answer:
            snooze_time = self.notification.snooze_time or 1
            self.next_display = datetime.datetime.now() + datetime.timedelta(days=snooze_time)
        super(UserNotification, self).save(*args, **kwargs)
        if self.notification.is_web:
            self.update_unseen_cache()
        elif self.notification.is_push:
            self.send_push_notification(lang)

    def delete(self, *args, **kwargs) -> None:
        """Override delete so it clears notification from cache as well"""
        super(UserNotification, self).delete(*args, **kwargs)
        cache.delete(Notification.notification_key(self.user))

    @staticmethod
    def update_for_user(user: User) -> None:
        """Update notifications for user

        Args:
            user (User): User instance
        """
        uns = user.usernotification_set.exclude(answer=True).filter(
            Q(notification__expires=None) | Q(notification__expires__gt=datetime.datetime.now())
        ).filter(notification__notification_type=NotificationAbstractModel.TypeChoices.WEB)

        # Here to ensure cache is up to date
        for un in uns:
            un.update_unseen_cache()


class Topic(models.Model):
    '''Model for Push notification topic in AWS SNS'''
    created_at = models.DateTimeField(auto_now_add=True)
    name = models.CharField(_('Topic name'), max_length=256)
    language = models.CharField(
        _('Language for topic'),
        choices=settings.LANGUAGES,
        max_length=6
    )
    arn = models.CharField(
        _('Topic ARN'),
        max_length=256,
        help_text=_('ARN used to identify topic in AWS SNS'),
        null=True,
        blank=True,
    )

    class Meta:
        verbose_name = _('Notification topic')
        verbose_name_plural = _('Notification topics')
        abstract = True

    def __str__(self) -> str:
        return self.name

    @staticmethod
    def get_subscription_model():
        raise NotImplementedError()

    def get_topic_name(self) -> str:
        """Method that defines how topic name will be referred to in AWS SNS
        Note that topic names must be unique

        Raises:
            NotImplemented: Not implemented for this abstract model

        Returns:
            str: Topic name as represented in AWS SNS
        """
        raise NotImplemented('You need to implement get_topic_name method')

    def save(self, **kwargs) -> 'Topic':
        """Override save to create topic in AWS SNS, if not created already"""
        if not self.arn:
            client = SNSClient()
            success, arn = client.create_topic(self.get_topic_name())
            if success:
                self.arn = arn
        return super().save(**kwargs)

    def delete(self, *args, **kwargs):
        """Override delete to delete from AWS as well"""
        if self.arn is not None:
            client = SNSClient()
            client.connection.delete_topic(TopicArn=self.arn)
        return super().delete(*args, **kwargs)

    @property
    def attributes(self) -> dict:
        """Get topic attributes from AWS SNS

        Returns:
            dict: Attributes retrieved from SNS
        """
        client = SNSClient()
        attributes = client.get_topic_attributes(self)
        return attributes

    @property
    def confirmed_subscriptions(self) -> int:
        return int(self.attributes.get('SubscriptionsConfirmed', 0))

    @property
    def pending_subscriptions(self) -> int:
        return int(self.attributes.get('SubscriptionsPending', 0))

    @property
    def deleted_subscriptions(self) -> int:
        return int(self.attributes.get('SubscriptionsDeleted', 0))

    def subscribe(self, device: 'Device') -> 'Subscription':
        """Subscribe device to topic

        Args:
            device (Device): Device instance

        Returns:
            Model: Subscription instance
        """
        client = SNSClient()
        try:
            response = client.connection.subscribe(
                TopicArn=self.arn,
                Endpoint=device.arn,
                Protocol='application',
                ReturnSubscriptionArn=True
            )
        except Exception:
            # Subscription failed, so return None
            return None
        SubModel = self.get_subscription_model()
        sub = SubModel(device_arn=device.arn, topic=self)
        sub.arn = response['SubscriptionArn']
        sub.save()
        return sub

    def unsubscribe(self, device: 'Device') -> None:
        """Unsubscribe device from topic

        Args:
            device (Device): Device instance
        """
        sub = self.subscriptions.filter(device_arn=device.arn).first()
        if sub is not None:
            sub.delete()

    def send(self, notification: NotificationAbstractModel) -> HTTPResponse:
        """Send notification to topic

        Args:
            notification (dict): Notification

        Raises:
            NotificationError: If notification could not be created in AWS

        Returns:
            HTTPResponse: Response from SNS
        """
        client = SNSClient()
        success, id_or_error = client.publish_to_topic(self, notification)
        if not success:
            raise NotificationError(f'Could not create notification. Response from AWS: {id_or_error}')
        with transaction.atomic():
            notification.aws_sns_id = id_or_error
            notification.save(update_fields=['aws_sns_id'])
        return notification

    def delete(self, **kwargs) -> Tuple[int, Dict[str, int]]:
        """Override delete so it also deletes topic from SNS

        Returns:
            Tuple[int, Dict[str, int]]: Delete results
        """
        client = SNSClient()
        try:
            client.connection.delete_topic(TopicArn=self.arn)
        except Exception:
            # Topic not found, so nothing to delete
            pass
        return super().delete(**kwargs)


class Subscription(models.Model):
    """Abstract model to handle AWS SNS topic subscriptions"""
    device_arn = models.CharField(
        _('Device arn'),
        max_length=255,
        help_text=_('ARN for device subscribed to topic'),
    )
    topic = models.ForeignKey(
        'NotificationTopic',
        related_name='subscriptions',
        on_delete=models.CASCADE,
    )
    arn = models.CharField(
        _('Subscription ARN'),
        max_length=255,
        help_text=_('Used to identify subscription in AWS')
    )
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        verbose_name = _('Device subscription')
        verbose_name = _('Device subscriptions')
        unique_together = ('device_arn', 'topic_id',)
        abstract = True

    def __str__(self) -> str:
        return f'{self.device_arn}: {self.topic}'

    @property
    def attributes(self) -> dict:
        """Get subscription attributes from AWS

        Returns:
            dict: Subscription attributes
        """
        client = SNSClient()
        try:
            return client.connection.get_subscription_attributes(
                SubscriptionArn=self.arn
            )
        except Exception:
            # Subsription doesn't exist in AWS, probably, so return empty dict
            return {}

    def delete(self, *args, **kwargs) -> None:
        """Override delete method so it will delete subscription from AWS as well"""
        client = SNSClient()
        try:
            client.connection.unsubscribe(
                SubscriptionArn=self.arn,
            )
        except Exception:
            # If error, then subscription probably doesn't exist, so nothing to do here
            pass
        return super().delete(*args, **kwargs)


class NotificationTopic(Topic):

    class Meta:
        verbose_name = _('Notification topic')
        verbose_name_plural = _('Notification topics')

    def get_topic_name(self) -> str:
        notif_identifier = get_tenant_identifier()
        return f'{notif_identifier}-{self.name}'

    @staticmethod
    def get_subscription_model():
        return DeviceSubscription


class SentToDevice(models.Model):
    """Model to track notifications sent to devices"""
    notification = models.ForeignKey('Notification', on_delete=models.CASCADE)
    device = models.ForeignKey('Device', on_delete=models.CASCADE)
    sent_at = models.DateTimeField(auto_now_add=True)
    message_id = models.UUIDField(
        _('Message id in AWS'),
        null=True,
        blank=True,
    )

    @property
    def sent(self):
        return self.message_id is not None


class Device(models.Model):
    """Model class for a user device that can receive push notifications"""
    class DeviceChoices(models.IntegerChoices):
        IOS = (0, 'IOS')
        ANDROID = (1, 'Android')

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    os = models.IntegerField(choices=DeviceChoices.choices)
    token = models.CharField(max_length=255, unique=True)
    arn = models.CharField(max_length=255, unique=True, blank=True, null=True)
    active = models.BooleanField(default=False)
    user = models.ForeignKey(
        User, on_delete=models.CASCADE, related_name='devices')
    sent_notifications = models.ManyToManyField(
        Notification,
        through=SentToDevice,
        through_fields=('device', 'notification'),
        related_name='sent_to_devices'
    )

    def __str__(self) -> str:
        return f'{self.user.username}: {self.os}'

    class Meta:
        verbose_name = _('Device')
        verbose_name_plural = _('Devices')

    @property
    def is_android(self) -> bool:
        return self.os == self.DeviceChoices.ANDROID

    @property
    def is_ios(self) -> bool:
        return self.os == self.DeviceChoices.IOS

    def delete(self, *args, **kwargs) -> Tuple[int, Dict[str, int]]:
        if self.arn is not None:
            self.deregister()
        return super().delete(*args, **kwargs)

    def update_from_aws(self) -> bool:
        """Method to update active status from what is registered in AWS

        Returns:
            bool: If updated
        """
        enabled = self.attributes.get('Enabled', None) == 'true'
        if self.active != enabled:
            self.active = enabled
            self.save(update_fields=['active'])
            return True
        return False

    def register(self) -> bool:
        """Method to register device to SNS so it can receive push notifications.
        It retrieves the arn for the device and saves it to use as the device
        identifier to send out push notifications

        Returns:
            bool: Successful registration
        """
        client = SNSClient()
        profile = getattr(self.user, 'profile', None)
        emp = getattr(profile, 'emp', None)
        empid = getattr(emp, 'empid', '')
        tenant = get_tenant_identifier()
        custom_user_data = f'{tenant}:empid({empid})'
        success, arn_or_error = client.create_platform_endpoint(
            self, custom_user_data=custom_user_data)
        if success:
            self.arn = arn_or_error
            self.active = True
            self.save(update_fields=['arn', 'active'])
        return success

    def deregister(self, update_instance: bool = True) -> None:
        """
        Method that deletes registered a device from SNS.
        """
        if self.arn is None:
            return
        client = SNSClient()
        client.delete_platform_endpoint(self)
        if update_instance:
            self.active = False
            self.save(update_fields=['active'])

    def send(self, notification: Notification, language: str = 'is') -> str:
        """Method that sends notification to device

        Args:
            title (str): Notification title
            text (str): Notification body
            notification_type (str): Notification type
            data (dict, optional): Notification data. Defaults to dict().
        Returns:
            HTTPResponse: Response from SNS
        """

        client = SNSClient()
        if SentToDevice.objects.filter(device=self, notification=notification).exists():
            raise NotificationError('Notification already sent to this device')
        success, id_or_error = client.publish_to_device(self, notification, language)
        with transaction.atomic():
            self.sent_notifications.through.objects.create(
                device=self,
                notification=notification,
                message_id=id_or_error if success else None,
            )
        if not success:
            raise NotificationError(id_or_error)

    def subscribe(self, topic: Topic) -> 'DeviceSubscription':
        """Subscribe device to topic

        Args:
            topic (Topic): Topic instance

        Returns:
            DeviceSubscription: Subscription instance
        """
        return topic.subscribe(self)

    def unsubscribe(self, topic: Topic) -> None:
        """Unsubscribe device from topic

        Args:
            topic (Topic): Topic instance
        """
        return topic.unsubscribe(self)

    @property
    def attributes(self) -> Dict[str, Any]:
        """Get attributes for device from AWS

        Returns:
            Dict[str, Any]: Attributes
        """
        try:
            client = SNSClient()
            return client.retrieve_platform_endpoint_attributs(self)
        except Exception:
            return {}


class DeviceSubscription(Subscription):
    SUB_TYPE = 'customer'



# -*- coding: utf-8 -*-
from rest_framework import serializers

from notifications.models import Notification, UserNotification


from django.contrib.auth import get_user_model
User = get_user_model()


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ('id', 'username', 'email', 'first_name', 'last_name', )


class NotificationSerializer(serializers.HyperlinkedModelSerializer):
    regex = serializers.CharField(source='display_only_if_url_path_matches_regex')
    class Meta:
        model = Notification
        fields = ('name', 'message', 'snooze_time', 'snooze_lock', 'look', 'image', 'regex',)


class UserNotificationSerializer(serializers.HyperlinkedModelSerializer):
    notification = NotificationSerializer()
    user = UserSerializer()

    class Meta:
        model = UserNotification
        fields = ('id', 'notification', 'user', 'seen', 'answer')


class UserNotificationPutSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = UserNotification
        fields = ('answer', 'answer_string')



# Generated by Django 3.2.13 on 2022-07-01 13:41

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import tinymce.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('notifications', '0004_auto_20200915_1222'),
    ]

    operations = [
        migrations.CreateModel(
            name='Device',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('os', models.IntegerField(choices=[(0, 'IOS'), (1, 'Android')])),
                ('token', models.CharField(max_length=255, unique=True)),
                ('arn', models.CharField(blank=True, max_length=255, null=True, unique=True)),
                ('active', models.BooleanField(default=False)),
            ],
            options={
                'verbose_name': 'Device',
                'verbose_name_plural': 'Devices',
            },
        ),
        migrations.CreateModel(
            name='NotificationTopic',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=256, verbose_name='Topic name')),
                ('language', models.CharField(choices=[('is', 'Íslenska'), ('en', 'English')], max_length=6, verbose_name='Language for topic')),
                ('arn', models.CharField(blank=True, help_text='ARN used to identify topic in AWS SNS', max_length=256, null=True, verbose_name='Topic ARN')),
            ],
            options={
                'verbose_name': 'Notification topic',
                'verbose_name_plural': 'Notification topics',
            },
        ),
        migrations.AlterModelOptions(
            name='notification',
            options={'verbose_name': 'Notification', 'verbose_name_plural': 'Notifications'},
        ),
        migrations.AlterModelOptions(
            name='usernotification',
            options={'verbose_name': 'User notification', 'verbose_name_plural': 'User notifications'},
        ),
        migrations.AddField(
            model_name='notification',
            name='aws_sns_id',
            field=models.UUIDField(blank=True, default=None, help_text='Message id in AWS, if notification is a push notification', null=True, verbose_name='Id of notification in AWS SNS'),
        ),
        migrations.AddField(
            model_name='notification',
            name='notification_type',
            field=models.IntegerField(choices=[(0, 'Web notification'), (1, 'Push notification'), (2, 'Email notification (not implemented)')], default=0, verbose_name='Notification type'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='attachment',
            field=models.FileField(blank=True, default=None, null=True, upload_to=''),
        ),
        migrations.AlterField(
            model_name='notification',
            name='expires',
            field=models.DateTimeField(blank=True, default=None, help_text='Notification will not be shown after this time.', null=True),
        ),
        migrations.AlterField(
            model_name='notification',
            name='look',
            field=models.CharField(choices=[('SIGN_COMPANY', 'Sign Company'), ('SIMPLE_OK', 'Simple Ok'), ('RELEASE_NOTES', 'Release Notes')], default='SIMPLE_OK', help_text='This controls the appearance of the notification.', max_length=50),
        ),
        migrations.AlterField(
            model_name='notification',
            name='message',
            field=tinymce.models.HTMLField(help_text='Full message as shown to user'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='name',
            field=models.CharField(help_text='Used to reference notification, shown in title for some looks', max_length=250, unique=True),
        ),
        migrations.AlterField(
            model_name='notification',
            name='needs_approval',
            field=models.BooleanField(default=False, help_text='Set this field if approval is necessary, see snooze time.'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='snooze_time',
            field=models.IntegerField(blank=True, default=None, help_text='If user dismisses message (when relevant), message is shown again after this many days.', null=True),
        ),
        migrations.AlterField(
            model_name='usernotification',
            name='answer',
            field=models.BooleanField(blank=True, default=None, null=True),
        ),
        migrations.CreateModel(
            name='SentToDevice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sent_at', models.DateTimeField(auto_now_add=True)),
                ('message_id', models.UUIDField(blank=True, null=True, verbose_name='Message id in AWS')),
                ('device', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='notifications.device')),
                ('notification', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='notifications.notification')),
            ],
        ),
        migrations.AddField(
            model_name='device',
            name='sent_notifications',
            field=models.ManyToManyField(related_name='sent_to_devices', through='notifications.SentToDevice', to='notifications.Notification'),
        ),
        migrations.AddField(
            model_name='device',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='devices', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='notification',
            name='topic',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='sent_notifications', to='notifications.notificationtopic'),
        ),
        migrations.CreateModel(
            name='DeviceSubscription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('device_arn', models.CharField(help_text='ARN for device subscribed to topic', max_length=255, verbose_name='Device arn')),
                ('arn', models.CharField(help_text='Used to identify subscription in AWS', max_length=255, verbose_name='Subscription ARN')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('topic', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions', to='notifications.notificationtopic')),
            ],
            options={
                'verbose_name': 'Device subscriptions',
                'abstract': False,
                'unique_together': {('device_arn', 'topic_id')},
            },
        ),
    ]



from django.conf import settings
from django.db import connection


def get_tenant_identifier() -> str:
    """Get name used to identify notification by tenant.
    Note that this is very specific to Timon needs.

    Returns:
        str: Tenant identifier
    """
    try:
        return connection.get_threadlocal().get_db_name()
    except Exception:
        return getattr(settings, 'WEB_NAME', '')



# -*- coding: utf-8 -*-
from datetime import datetime
from typing import Any, Optional, Sequence

from django.contrib import admin
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from .models import (
    Device,
    DeviceSubscription,
    Notification,
    NotificationTopic,
    SentToDevice,
    UserNotification
)


@admin.register(UserNotification)
class UserNotificationAdmin(admin.ModelAdmin):
    actions = ['really_delete_selected']

    list_display = (
        'user',
        'notification'
    )

    search_fields = (
        'user__username',
        'notification__name'
    )
    ordering = ('user__username', 'notification__name')

    def get_actions(self, request):
        actions = super(UserNotificationAdmin, self).get_actions(request)
        del actions['delete_selected']
        return actions

    @admin.action(description="Delete selected entries")
    def really_delete_selected(self, request, queryset):
        for obj in queryset:
            obj.delete()

        # The queryset cache has the count after the loop above, no
        # need for extra call or storing count in variable.
        if queryset.count() == 1:
            message_bit = "1 UserNotification entry was"
        else:
            message_bit = "%s usernotification entries were" % queryset.count()
        self.message_user(request, "%s successfully deleted." % message_bit)


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    readonly_fields = ('management', 'aws_sns_id')
    fields = (
        'name_en',
        'name_is',
        'message_en',
        'message_is',
        'notification_type',
        'active_from',
        'expires',
        'needs_approval',
        'snooze_time',
        'groups',
        'look',
        'image',
        'management',
        'aws_sns_id',
        'topic',
    )
    list_display = (
        'name',
        'notification_type',
        'expires',
        'look',
        'needs_approval',
    )
    ordering = ('-id', )

    def has_change_permission(self, request: HttpRequest, obj: Optional[Notification] = None) -> Sequence[str]:
        if obj is not None and obj.aws_sns_id:
            return False
        return super().has_change_permission(request, obj)

    @admin.display(description='Management')
    def management(self, instance):
        EXCLUDED_FIELDS = [
            'groups',
            'id',
            'recipients',
            'attachment',
            'active_from',
            'snooze_lock',
            'recipients',
            'send_email',
            'aws_sns_id'
        ]
        params = ''
        for field in instance.__class__._meta.get_fields():
            if field.name in EXCLUDED_FIELDS:
                continue
            attr = getattr(instance, field.name, None)
            if attr is None or attr == '':
                continue
            if isinstance(attr, datetime):
                attr = attr.isoformat()
            if field.name == 'message':
                attr = ''.join(attr.split('\n'))
                attr = ''.join(attr.split('\r'))
            params += ' --{}=\'{}\''.format(field.name, attr)
        groups = ','.join(map(str, instance.groups.values_list('id', flat=True)))
        if groups:
            params += ' --groups={}'.format(groups)

        return "python manage.py import_notification{}".format(params)

    def save_model(self, request, obj, form, change):
        super(NotificationAdmin, self).save_model(request, obj, form, change)
        for group in form.cleaned_data['groups']:
            for user in group.user_set.all():
                UserNotification.objects.get_or_create(notification=obj, user=user)


class SentInline(admin.TabularInline):
    model = SentToDevice
    fields = ('notification', 'sent_at', 'get_sent', 'message_id')
    readonly_fields = ('notification', 'sent_at', 'get_sent', 'message_id')
    extra = 0
    can_delete = False

    def get_max_num(self, request: HttpRequest, obj: Optional[Notification] = None, **kwargs: Any) -> Optional[int]:
        if obj is not None:
            return obj.sent_notifications.count()
        return super().get_max_num(request, obj, **kwargs)

    def has_change_permission(self, *args, **kwargs) -> bool:
        return False

    @admin.display(description='Successfully sent')
    def get_sent(self, obj):
        return obj.sent


@admin.register(Device)
class DeviceAdmin(admin.ModelAdmin):
    inlines = [SentInline]
    list_display = (
        'user',
        'os',
        'token',
        'arn',
        'active',
        'created_at',
    )
    readonly_fields = (
        'arn',
        'active',
        'created_at',
        'updated_at',
    )

    def get_actions(self, request):
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions


class NotificationInline(admin.TabularInline):
    model = Notification
    fields = ('name', 'message')
    extra = 0
    can_delete = False
    show_change_link = True

    def get_max_num(self, request: HttpRequest, obj: Optional[Notification] = None, **kwargs: Any) -> Optional[int]:
        if obj is not None:
            return obj.sent_notifications.count()
        return super().get_max_num(request, obj, **kwargs)

    def has_change_permission(self, *args, **kwargs) -> bool:
        return False


@admin.register(NotificationTopic)
class NotificationTopicAdmin(admin.ModelAdmin):
    inlines = [NotificationInline]
    list_display = (
        'name', 'language', 'arn', 'subscribed', 'sent'
    )
    readonly_fields = ('arn',)

    def get_actions(self, request):
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions

    @admin.display(description='Subscribed')
    def subscribed(self, obj):
        return obj.subscriptions.count()

    @admin.display(description='Sent notifications')
    def sent(self, obj):
        return obj.sent_notifications.count()


@admin.register(DeviceSubscription)
class DeviceSubscriptionAdmin(admin.ModelAdmin):
    list_display = ('device_arn', 'topic', 'arn',)
    readonly_fields = ('device_arn', 'topic', 'arn',)



from http.client import HTTPResponse
from typing import Optional, Tuple
from django.conf import settings
from django.db.models import Model
from django.utils.translation import activate as activate_translation
from botocore.client import BaseClient
from botocore.exceptions import ClientError
import boto3
import json

from django.http import HttpResponse


class SNSClient(object):
    """ Class representing an AWS SNS client that supports mobile push notifications.
    """

    def __init__(self) -> None:
        self.connection = self.connect()
        # retrieve AWS credentials from settings.
        self.ios_arn = getattr(settings, 'IOS_PLATFORM_APPLICATION_ARN')
        self.android_arn = getattr(
            settings, 'ANDROID_PLATFORM_APPLICATION_ARN')

    @staticmethod
    def connect() -> BaseClient:
        """Method that creates a connection to AWS SNS

        Raises:
            NotImplementedError: If AWS settings are not defined in settings

        Returns:
            BaseClient: AWS client instance
        """
        # start an AWS session.
        session = boto3.Session()
        if getattr(settings, 'AWS_SNS_REGION_NAME', None) and getattr(settings, 'AWS_ACCESS_KEY_ID', None):
            return session.client(
                "sns",
                region_name=getattr(settings, 'AWS_SNS_REGION_NAME'),
                aws_access_key_id=getattr(settings, 'AWS_ACCESS_KEY_ID'),
                aws_secret_access_key=getattr(
                    settings, 'AWS_SECRET_ACCESS_KEY'),
            )
        raise NotImplementedError('AWS settings are missing in environment')

    def retrieve_platform_endpoint_attributs(self, device: Model) -> dict:
        """Method that retrieves endpoint attributes for a device

        Args:
            device (Device): Device instance

        Returns:
            dict: Endpoint attributes
        """
        try:
            response = self.connection.get_endpoint_attributes(
                EndpointArn=device.arn
            )
            return response['Attributes']
        except ClientError:
            return {}

    def delete_platform_endpoint(self, device: Model) -> bool:
        """Deletes endpoint for given device

        Args:
            device (Device): Device instance

        Returns:
            bool: If successfully deleted
        """
        try:
            self.connection.delete_endpoint(
                EndpointArn=device.arn
            )
            return True
        except ClientError:
            return False

    def create_platform_endpoint(self, device: Model, custom_user_data: Optional[str] = None) -> Tuple[bool, str]:
        """Method that creates a platform endpoint for an Android device.

        Args:
            device (Device): Device instance
            custom_user_data (str): String to tie to user data

        Returns:
            Tuple[bool, str]: Success and arn or error message
        """
        arn = None
        if device.is_ios:
            arn = self.ios_arn
        elif device.is_android:
            arn = self.android_arn
        try:
            response = self.connection.create_platform_endpoint(
                PlatformApplicationArn=arn,
                Token=device.token,
                CustomUserData=custom_user_data
            )
            return True, response['EndpointArn']
        except ClientError as e:
            error_resp = e.response['Error']
            return False, error_resp['Message']

    def format_ios_message(
        self,
        title: str,
        text: str,
        notification_id: int = None
    ) -> dict:
        """Formats a push notification to be compatible with IOS notifications

        Args:
            title (str): Title of notification
            text (str): Contents of notification
            notification_type (str): Type of notification
            data (dict): Notification data
            notification_id (int): Notification id

        Returns:
            dict: Formatted message
        """
        return json.dumps({
            'aps': {
                'alert': {
                    'title': title,
                    'body': text,
                },
            },
            'id': notification_id,
        })

    def format_android_message(
        self,
        title: str,
        text: str,
        notification_id: int = None
    ) -> dict:
        """Formats a push notification to be compatible with Android notifications

        Args:
            title (str): Title of notification
            text (str): Contents of notification
            notification_type (str): Type of notification
            data (dict): Notification data
            notification_id (int): Notification id

        Returns:
            dict: Formatted message
        """
        return json.dumps({
            'notification': {
                'title': title,
                'text': text,
                'body': text,
            },
            'data': {
                'id': notification_id,
            }
        })

    def publish_to_device(
        self,
        device: Model,
        notification: Model,
        language: Optional[str] = None
    ) -> Tuple[str, HttpResponse]:
        """Publish a push notification to an endpoint

        Args:
            device (Device): Device instance
            notification (Notification): Notification instance
            language (str): Language for device

        Returns:
            Tuple[str, str]: Tuple containing published message along with the response from AWS
        """
        if notification is None:
            return False, "No notification to publish"
        if language is None:
            default_language = getattr(settings, 'LANGUAGE_CODE', 'en')
            profile = getattr(self.user, 'profile', None)
            prefs = getattr(profile, 'prefs', {})
            language = prefs.get('language', default_language)
        activate_translation(language)
        notification_id = notification.id
        title = notification.name
        message = {'default': notification.message}
        if not device.active:
            return False, 'Device is inactive'
        if device.is_ios:
            message['APNS'] = self.format_ios_message(title, message, notification_id)
        elif device.is_android:
            message['GCM'] = self.format_android_message(title, message, notification_id)
        try:
            response = self.connection.publish(
                TargetArn=device.arn,
                Message=json.dumps(message),
                MessageStructure='json',
            )
            return True, response['MessageId']
        except ClientError as e:
            error_resp = e.response['Error']
            message = error_resp['Message']
            if error_resp['Code'] == 'EndpointDisabled':
                device.active = False
                device.save(update_fields=['active'])
                message = f'{message}: Device marked as inactive'
            return False, message

    def publish_to_topic(self, topic: Model, notification: Model) -> Tuple[bool, str]:
        """Publish a push notification to a topic

        Args:
            topic_arn (Topic): Topic instance
            notification (Notification): Notification instance

        Returns:
            Tuple[bool, str]: Whether successful and then id or error message
        """
        activate_translation(topic.language)
        try:
            response = self.connection.publish(
                TopicArn=topic.arn,
                Subject=notification.name,
                Message=notification.message,
            )
            return True, response['MessageId']
        except ClientError as e:
            error_resp = e.response['Error']
            message = error_resp['Message']
        return False, message

    def create_topic(self, topic_name: str) -> Tuple[bool, str]:
        """Create topic in AWS

        Args:
            topic_name (str): Topic name

        Returns:
            Tuple[bool, str]: Whether successful and then topic arn or error message
        """
        try:
            response = self.connection.create_topic(Name=topic_name)
            return True, response['TopicArn']
        except ClientError as e:
            error_resp = e.response['Error']
            message = error_resp['Message']
            return False, message

    def get_topic_attributes(self, topic: Model) -> dict:
        """Gets attributes for topic

        Args:
            topic (Model): Topic instance

        Returns:
            dict: Topic attributes
        """
        try:
            return self.connection.get_topic_attributes(TopicArn=topic.arn)['Attributes']
        except ClientError:
            return {}



# -*- coding: utf-8 -*-
from urllib.parse import urlparse
import datetime
import operator
import re

from .models import Notification, UserNotification




class NotificationMiddleware(object):
    IGNORED_PATHS = ['jsi18n']

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        if self.is_valid_request(request):
            request.notifications = Notification.unseen(request.user)
            response = self.process_response(request, response)
        return response

    def is_valid_request(self, request):
        """Checks if request is valid
        Checks if
            - user is authenticated
            - request path is not in IGNORED_PATHS

        Args:
            request (Request): Request object

        Returns:
            bool: Whether request is valid
        """
        path_split = request.path.split('/')
        is_valid_path = (
            len(path_split) > 1 and path_split[1] not in self.IGNORED_PATHS
        ) or len(path_split) <= 1
        return hasattr(request, 'user') and request.user.is_authenticated and is_valid_path

    def process_response(self, request, response):
        """Process response and add notification values to cookies if appliccable

        Args:
            request (Request): Request instance
            response (Response): Response instanve

        Returns:
            Response: Response with added cookie, if appliccable
        """
        if not self.is_valid_request(request):
            return response

        notifications = Notification.unseen(request.user)
        now = datetime.datetime.now()
        if notifications:
            notifications_keys = notifications.keys()
            notification_value_list = UserNotification.objects.filter(id__in=notifications_keys).values_list(
                'id', 'notification__display_only_if_url_path_matches_regex', 'notification__active_from',
                'notification__expires')
            for notif_id, regex, active_from, expires in notification_value_list:
                if active_from is not None and now < active_from:
                    del notifications[notif_id]
                    continue
                if expires is not None and now > expires:
                    del notifications[notif_id]
                    continue
                matching_path = re.search(regex, request.path) is not None
                if not matching_path and notif_id in notifications:
                    del notifications[notif_id]

        if notifications:
            # Need to ensure cookie string is in reverse order from closest to last
            sorted_items = sorted(notifications.items(), key=operator.itemgetter(1))
            epoch = datetime.datetime.utcfromtimestamp(0)
            max_age = 14 * 24 * 60 * 60
            expires = datetime.datetime.utcnow() + datetime.timedelta(seconds=max_age)

            coookie_string = "-".join(
                '{}:{}'.format(key, (val - epoch).total_seconds() * 1000.0)
                for key, val in sorted_items
            )
            response.set_cookie(
                'notifications', coookie_string, expires=expires
            )
        else:
            response.delete_cookie('notifications')

        return response



import datetime
import json

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django_redis import get_redis_connection

from notifications.models import Notification, UserNotification
from timonutils.tests.base import BaseClientTest

User = get_user_model()


class NotificationTest(BaseClientTest):
    SNOOZE_TIME = 1

    def setUp(self):
        self.tearDown()
        super(NotificationTest, self).setUp()
        self.notification = Notification.objects.create(
            name_en="test",
            name_is="test",
            message_en="test message",
            message_is="test message",
            snooze_time=self.SNOOZE_TIME
        )
        self.user = User.objects.get(username=self.USERNAME)
        self.un = UserNotification.objects.create(notification=self.notification, user=self.user)
        self.created = datetime.datetime.now()

    def tearDown(self):
        get_redis_connection("default").flushall()

    def test_creation(self):
        self.assertEqual(len(Notification.unseen(self.user)), 1)

    def test_seen(self):
        self.un.seen = True
        self.un.answer = True
        self.un.save()
        self.assertEqual(len(Notification.unseen(self.user)), 0)
        self.un.answer = False
        self.un.save()
        self.assertEqual(len(Notification.unseen(self.user)), 1)

    def test_cookie(self):
        response = self.client.get('/api/')
        notification_cookie = response.cookies['notifications'].value.split(':')
        unseenid = notification_cookie[0]
        self.assertEqual(int(unseenid), self.un.id)
        unseentime = float(notification_cookie[1]) / 1000

        utcfromtimestamp = datetime.datetime.utcfromtimestamp(unseentime)
        self.assertTrue(abs(utcfromtimestamp - self.created) < datetime.timedelta(seconds=5, minutes=10))

        self.un.seen = True
        self.un.answer = False
        self.un.save()
        snooze = datetime.datetime.now() + datetime.timedelta(days=self.SNOOZE_TIME)
        response = self.client.get('/api/')
        notification_cookie = response.cookies['notifications'].value.split(':')
        unseenid = notification_cookie[0]
        self.assertEqual(int(unseenid), self.un.id)
        unseentime = float(notification_cookie[1]) / 1000
        self.assertTrue(abs(datetime.datetime.utcfromtimestamp(unseentime) - snooze) < datetime.timedelta(seconds=1))

    def test_expired(self):
        response = self.client.get('/api2/usernotifications/')
        self.assertEqual(len(json.loads(response.content)), 1)
        self.notification.expires = datetime.datetime.now() - datetime.timedelta(1)
        self.notification.save()

        response = self.client.get('/api2/usernotifications/')
        self.assertEqual(len(json.loads(response.content)), 0)

        # Make sure model interface returns same as api
        self.assertEqual(len(Notification.unseen(self.user)), 0)

    def test_deletion(self):
        self.un.delete()
        self.assertEqual(len(Notification.unseen(self.user)), 0)

    def test_group_creation(self):
        self.un.delete()
        self.assertEqual(len(Notification.unseen(self.user)), 0)
        group = Group.objects.last()
        self.user.groups.add(group)
        self.notification.groups.add(group)
        self.notification.create_usernotifications_for_groups()
        self.assertEqual(len(Notification.unseen(self.user)), 1)

        # No duplicates
        self.notification.groups.add(group)
        self.assertEqual(len(Notification.unseen(self.user)), 1)



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='trackwell-notifications',
    version='4.0.0',
    packages=find_packages(),
    include_package_data=True,
    license='NA',  # example license
    description='A simple Django app to support Notifications.',
    long_description=README,
    long_description_content_type='text/markdown',
    url='',
    author='trackwell',
    author_email='',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'Django', 'djangorestframework', 'django-filter',
    ]
)
